-- name=dbg_match3

--------------------------------------------------------------------------------
---- function *** Init *** (
--------------------------------------------------------------------------------
function public.__InitModule()
  private.max_sections = 0;
  private.current_section = 0;
  private.match3 = {};
  
  ne.msg.Subscribe( Event_Match3_FieldActivated, private.Match3_FieldActivated );
  ne.msg.Subscribe( "Event_Cheat_KeyPressed", private.Handle_Event_Cheat_KeyPressed );
end;
--------------------------------------------------------------------------------
function public.__InitObjects()
  ObjAttach("obj_dbg_match3", "ng_application");
  private.Init_Settings();
end;
--------------------------------------------------------------------------------
function public.Start()
  math.randomseed( os.time() );
  
  private.LoadStandardCursors();
  
  ne.msg.Subscribe( Event_Application_LoadingComplete, private.AppLoadingComplete );

  ne.msg.Send( Event_Menu_LoadingComplete, {} );
end;
--------------------------------------------------------------------------------
function private.LoadStandardCursors()
  SetSysCursor( GetSysCursor() );
  local count = 0;
  local CursorLoad = function ( cur_name, cur_path, cur_idx )
    _G[ cur_name ] = count;
    CursorLoad( cur_path, cur_idx );
    count = count + 1;
  end;

  CursorLoad( "CURSOR_NULL",    "", 0 );
  CursorLoad( "CURSOR_DEFAULT", "assets/cursors/cursor_default_a", 2 );
  CursorLoad( "CURSOR_HAND",    "assets/cursors/cursor_hand_a",    4 );

  SetCursor( CURSOR_DEFAULT );
end;
--------------------------------------------------------------------------------
function private.Info( text )
  DbgTrace("[INFO] "..tostring(text));
end;
--------------------------------------------------------------------------------
function private.Exception( text )
  DbgTrace("[ERROR] "..tostring(text));
end;
--------------------------------------------------------------------------------
function private.AppLoadingComplete()
  private.Info( "Loaded!" );
  
  parts.db.RegisterReadonlyCollections( { "assets/content/database/logic_settings" } );
  local sections_count_query = {
    query      = "find",
    collection = "logic_settings:type",
    criteria = { like = "sections_enable" }
  };
  private.max_sections = tonumber(parts.db.ExecuteQuery(sections_count_query).result.value);

end;
--------------------------------------------------------------------------------
function private.Handle_Event_Cheat_KeyPressed(msg, params)
  local key    = params.key or 0;
  if (key == 49 ) then
    ne.msg.Send( Event_Match3_BoosterActivated,
    {
      name     = "hammer",
      row      = nil,
      col      = nil,
    vertical = nil });
  end;
end;
--------------------------------------------------------------------------------
function private.TryLoadSection()
  local section = private.current_section + 1;
  if ( section > private.max_sections ) then
    private.Exception( "TryLoadSection: requested section <"..tostring(section).."> > max section <"..tostring(private.max_sections)..">" );
    return false;
  end;
  local section_name = "stages_"..tostring(section);
  parts.db.RegisterReadonlyCollections( { "assets/content/levels/"..section_name } );
  private.current_section = section;
  private.Info("Section <"..tostring(private.current_section ).."> loaded.");
  return true;
end;
--------------------------------------------------------------------------------
-- function *** Match3 *** ()
--------------------------------------------------------------------------------
function private.Match3_Refresh()
  ObjDelete("match");
  
  ObjCreate( "match", "matchbox" );
  ObjSet( "match", 
  { 
    stage     = "stage_"..tostring(private.match3.stage), 
    state     = private.match3.state,
    boosters  = {},
    section   = private.match3.section,
    attempts  = 0,
    max_attempts  = 999
  } );
  
  ObjAttach( "match", "obj_matchbox" );
end;
--------------------------------------------------------------------------------
function private.Match3_FieldActivated()
  private.match3.moves = private.match3.moves - 1;
  private.Match3_RefreshInfo();
end;
--------------------------------------------------------------------------------
function private.Match3_RefreshInfo()
  ObjSet( "txt_moves", { text = private.match3.moves } );
end;
--------------------------------------------------------------------------------
-- function *** Settings *** ()
--------------------------------------------------------------------------------
function private.Init_Settings()
  ObjSet( "spr_stage_back", { event_mdown = function() ObjSet( "ted_stage", { input = true } ); end } );
  ObjSet( "ted_stage", { text = "1" } );
  
  ObjSet( "spr_state_back", { event_mdown = function() ObjSet( "ted_state", { input = true } ); end } );
  ObjSet( "ted_state", { text = "1" } );
  
  
  ObjSet( "obj_lost_input", { event_mdown = private.OnLostFocus } );
  ObjSet( "btn_set", { event_mup = private.Set } );
end;
--------------------------------------------------------------------------------
function private.OnLostFocus()
  ObjSet( "ted_stage", { input = false } );
  ObjSet( "ted_state", { input = false } );
end;
--------------------------------------------------------------------------------
function private.Set()
  ObjSet( "ng_application", { input = false } );
  
  local state = tonumber(ObjGet("ted_state").text);
  local stage = tonumber(ObjGet("ted_stage").text);
  
  local test_result = nil;
  repeat 
    local test_load_query = {
      query      = "find",
      collection = "stages:name",
      criteria = { like = "stage_"..tostring(stage) }
    };
    test_result = parts.db.ExecuteQuery(test_load_query).result;

  until ( test_result or not private.TryLoadSection() );
  
  if ( not test_result ) then
    private.Exception( "Set: Error loading stage <"..tostring(stage)..">" );
  else
    private.match3 = {};
    
    private.match3.state      = state;
    private.match3.stage      = stage;
    private.match3.section    = test_result.section;
    
    local stateDb             = test_result.state[state];
    private.match3.moves      = stateDb.moves;
    
    private.Match3_Refresh();
    private.Match3_RefreshInfo();
  end;
  
  
  ObjSet( "ng_application", { input = true } );
end;
