-- name=int_dialog_offers
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Close      = function () end,
  Buy        = function () end,
  GetLotInfo = function () end
};
local _this = {};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  _this = {
    int      = "int_dialog_offers",
    dialog   = "obj_int_dialog_offers",
    window   = "anm_int_dialog_offers_window",
    button = {
      close = "btn_int_dialog_offers_close",
      soft  = "btn_int_dialog_offers_buy_soft",
      premium = "btn_int_dialog_offers_buy_premium"
    },
    rewards = {
      reward_1 = "anm_int_dialog_offers_reward_1",
      reward_2 = "anm_int_dialog_offers_reward_2",
      reward_3 = "anm_int_dialog_offers_reward_3"
    }
  };

  private.app_o = ui.screen.Orientation();
  private.app_h = ui.screen.Height();
  private.app_w = ui.screen.Width();
  private.window_area = {
    landscape = { -230, -180, 460, 510 },
    portrait = { -290, -210, 580, 620 };
  };

  private.slots = {
    center_slot = 1,
    left_slot = 2,
    right_slot = 3
  };
  
  private.Adapt();
  private.CreateButtons();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Show ( offer, type_offer, price )
  
  private.SetDefault( offer, type_offer, price );
  
  private.type_offer = type_offer;
  
  ui.window.Open( "offers", function ()
    
    ui.anim.Play(
      _this.window,
      "appear_"..private.type_offer.."_"..private.app_o,
      nil,
      "static_"..private.type_offer.."_"..private.app_o
    );
    
  end );
  
end;
--------------------------------------------------------------------------------
function public.Hide ( func_OnHide )
  
  ui.anim.Play(
    _this.window,
    "disappear_"..private.type_offer.."_"..private.app_o,
    function ()
      ui.window.Close( "offers", func_OnHide );
    end
  );
  
end;
--------------------------------------------------------------------------------
function public.ForceHide ()
  
  _f.Close();
  
end; 
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()
  
  local multiset = {};
  
  ui.obj.ToCenter( _this.dialog, 0, 0, multiset );
  ui.window.SetArea( _this.dialog, private.window_area[ private.app_o ], multiset );
  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.CreateButtons ()
  
  ui.button.Create(
    _this.button.close,
    {
      design = "close"
    },
    function() _f.Close(); end
  );
  
  ui.button.Create(
    _this.button.premium,
    {
      design = "buy",
      timer  = ""
    },
    function() _f.Buy( "premium" ); end
  );
  
  ui.button.Create(
    _this.button.soft,
    {
      title  = "",
      design = "price"
    },
    function() _f.Buy( "soft" ); end
  );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.SetDefault ( offer, type_offer, price )
  
  local multiset = {};
  
  ui.obj.Set( _this.window, 
  { 
    tt_title = "str_int_dialog_offers_"..type_offer.."_title",
    tt_description = "str_int_dialog_offers_"..type_offer.."_description"
  }, multiset );
  
  if ( type_offer == "soft" ) then
    ui.obj.Disable( _this.button.premium );
    ui.button.SetText( _this.button.soft, price.count, multiset );
  else
    ui.obj.Disable( _this.button.soft );
    local price_string = _f.GetLotInfo( price.lot_id ).price;
    price_string = tern( price_string == "" )( "str_int_dialog_offers_premium_buy", price_string );
    ui.button.SetText( _this.button.premium, price_string );
  end;
  
  ui.obj.Enable( _this.button[ type_offer ] );
  
  local table_rewards_count = {};
  
  for slot, reward in pairs( offer ) do
    
    local id = private.slots[ slot ];
    
    table_rewards_count[ "tt_count_"..id ] = reward.count;
    
    ui.obj.Set( _this.rewards[ "reward_"..id ], 
    { 
      animfunc = "reward_"..reward.type,
      playing = true,
      rt_reward = ui.obj.Res( reward.type, reward.name )
    } );
    
  end;
  
  ui.obj.Set( _this.window, table_rewards_count, multiset );
  
  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------