-- name=int_dialog_world_open
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Play    = function () end,
  Close   = function () end,
  OnClose = function () end
};

local _this = {};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()
  
  _this.dialog = "obj_int_dialog_world_open";
  _this.header = "anm_int_dialog_world_open_header";
  _this.window = "anm_int_dialog_world_open_window";
  _this.icon   = "anm_int_dialog_world_open_icon";

  _this.button = {
    play  = "btn_int_dialog_world_open_play",
    close = "btn_int_dialog_world_open_close"
  };

  _this.app_o = ui.screen.Orientation();

  private.Adapt();
  private.CreateButtons();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show ( section_number, section_name, section_progress )

  if ( _this.is_active ) then return; end;

  _this.is_active = true;

  private.SetDefault( section_number, section_name, section_progress );

  ui.anim.Play(
    _this.header,
    "appear_".._this.app_o,
    nil,
    "static_".._this.app_o
  );

  ui.window.Open( "world_open" );

end;
--------------------------------------------------------------------------------
function public.Hide ( func_end )

  if ( not _this.is_active ) then return; end;
  
  _this.is_active = false;

  _f.Close();

  ui.window.Close( "world_open", function () _f.OnClose( func_end ); end );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()

  local multiset    = {};
  local window_area = {
    landscape = { -165, -260, 330, 424 },
    portrait  = { -204, -317, 408, 520 }
  };

  ui.obj.ToCenter( _this.dialog, 0, 0, multiset );
  ui.window.SetArea( _this.dialog, window_area[ _this.app_o ], multiset );
  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.CreateButtons ()

  ui.button.Create(
    _this.button.close,
    {
      design = "close"
    },
    function () public.Hide(); end
  );

  ui.button.Create(
    _this.button.play,
    {
      title  = "str_int_dialog_world_open_play",
      design = "medium",
      color  = "green"
    },
    function() _f.Play(); end
  );

end;
--------------------------------------------------------------------------------
function private.SetDefault ( section_number, section_name, section_progress )

  local title    = ne.strings.Get( "str_int_dialog_worldmap_"..section_name );
  local multiset = {};
  local pattern  = "%s/%s";
  local bg_path  = "assets/content/"..section_name.."/dialog_bg";
  
  local tt_progress = pattern:format(
    section_progress.done,
    section_progress.total
  );

  ui.obj.Set(
    _this.window,
    {
      tt_title    = title,
      tt_progress = tt_progress,
      rt_back     = bg_path,
      playing     = true
    },
    multiset
  );

  local folder = "";
  if ( not section_name:match( "section_%d+" ) ) then
    folder = "/"..section_name;
  end;
  
  ui.obj.Set(
    _this.icon,
    {
      res = "assets/content"..folder.."/globalmap/anm_icon",
      animfunc = "icon"..section_number
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------