-- name=int_panel_questlist
--------------------------------------------------------------------------------
-- function DelegateScheme: ()
--------------------------------------------------------------------------------
local _f =
{
  ViewPort_Restore  = function ( anm_time )        end,
  ViewPort_Contract = function ( width, anm_time ) end,
  Interface_Show    = function ()                  end,
  Interface_Hide    = function ()                  end,
  Highlight_Show    = function ( focus_data )      end, -- list_item.focus
  Highlight_Hide    = function ()                  end,
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  private.obj =
  {
    interface   = "int_panel_questlist",
    questlist   = "anm_int_panel_questlist",
    scroller    = "scc_int_panel_questlist",
    transparent = "obj_int_panel_questlist_transparent",
    cell_item   = "obj_int_panel_questlist_cell",
    cell_anim   = "anm_int_panel_questlist_cell",
    cell_dummy  = "obj_int_panel_questlist_dummy",
    quest_anim  = "anm_int_panel_questlist_quest",
    target_spr  = "spr_int_panel_questlist_quest_icon",
    target_anm  = "anm_int_panel_questlist_quest_icon",
    target_txt  = "txt_int_panel_questlist_quest_num",
    reward_anm  = "anm_int_panel_questlist_reward",
    reward_icon_spr  = "spr_int_panel_questlist_reward_icon",
    reward_icon_anm  = "anm_int_panel_questlist_reward_icon"
  };
  
  private.strings =
  {
    quests     = "str_int_panel_questlist_title",
    no_quests  = "str_int_panel_questlist_no_quests"
  };
  
  private.offset    = 45;
  private.scc_h     = ui.screen.Height() - private.offset;
  private.cellsize  = 116;
  private.cells_max = math.floor( private.scc_h / private.cellsize );
  
  private.panel =
  {
    width     = 320,
    time_open = 0.3,
    time_hide = 0.3
  };
  
  private.frame =
  {
    expand   = 120,
    contract = 100
  }

  private.panel_w   = 320;
  private.anim_time = 0.3;
  
  private.questlist     = {};
  private.selected      = nil;
  private.dummies_count = 0;
  
--[[
  local list_item =
  {
    name = "quest_name",
    priority = 10,
    target =
    { 
      type    = "enemy",
      name    = "bandit",
      content = "section_1", -- optional
      stage   = "stage_101"  -- optional
    },
    focus =
    {
      targets = 
      {
        { 
          type = "fight",
          name = "bandit", 
          stage = "stage_101" -- optional
        }
      },
      current = 1
    },
    
    -- Optional fields:
    reward =
    { 
      type    = "fragment",
      name    = "fragment_42"
    },
    progress =
    {
      current = 5,
      total   = 10
    }
  };
]]
  
end;
--------------------------------------------------------------------------------
function public.__InitObjects ()
  
  local screen_w = ui.screen.Width();
  local screen_h = ui.screen.Height();
  
  ui.hub.Create( private.obj.interface );
  
  local multiset = {};
  
  ui.obj.Set(
    private.obj.interface,
    {
      pos_x = screen_w, 
      pos_y = 0 
    },
    multiset
  );
  
  ui.obj.Set(
    private.obj.scroller,
    {
      cellsize    = private.cellsize,
      inputrect_y = private.offset,
      inputrect_h = private.scc_h,
      croprect_y  = private.offset,
      croprect_h  = private.scc_h,
      enddragtrig = private.EndDragTrig
    },
    multiset
  );
  
  ui.obj.Set(
    private.obj.transparent,
    {
      pos_x          = -screen_w,
      inputrect_init = true,
      inputrect_w    = screen_w,
      inputrect_h    = screen_h,
      event_mup    = function () public.Hide(); end
    },
    multiset
  );
  
  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Show ()

  if ( private.active ) then return; end;
  private.active = true;

  ui.hub.AttachContent( private.obj.interface );
  
  ui.input.Off( private.obj.questlist );
  ui.obj.Set( private.obj.scroller, { scrollpos = 0 } );
   common.sound.Play( "quest_out" );

  _f.Interface_Hide();
  _f.ViewPort_Contract( private.panel.width, private.panel.time_open );
  
  ui.anim.Play(
    private.obj.questlist,
    "open",
    function ()
    
      ui.input.On( private.obj.questlist );
      mediator.Notify( "View_Questlist_Opened", private.questlist );
    
    end,
    "static"
  );

end;
--------------------------------------------------------------------------------
function public.Hide ()

  if ( not private.active ) then return; end;
  private.active = false;

  ui.input.Off( private.obj.questlist );
   common.sound.Play( "quest_in" );
  
  _f.ViewPort_Restore( private.panel.time_open );
  _f.Interface_Show();
  
  ui.anim.Play(
    private.obj.questlist,
    "close",
    function ()
    
      ui.input.On( private.obj.questlist );
      ui.hub.DetachContent( private.obj.interface );
    
    end
  );
  
  v_log.Playhaven_LogPlacement( "quest_tab_close" );
  
  if ( private.selected ) then
  
    private.Selection_Off( private.selected );
    
  end;

end;
--------------------------------------------------------------------------------
function public.IsActive ()

  return private.active;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.SetList ( list )

  for id = #private.questlist, 1, -1 do
  
    private.Item_Delete( id );
  
  end;
  
  local multiset = {};
  
  for _, quest_data in pairs( list ) do
  
    private.Item_Create( quest_data, multiset );
  
  end;
  
  if ( #multiset > 0 ) then ui.obj.Multiset( multiset ); end;
  
  private.Dummies_Set();
  private.Title_Set();

end;
--------------------------------------------------------------------------------
function public.Add ( quest_data )

  local multiset = {};

  private.Item_Create( quest_data, multiset );
  
  if ( #multiset > 0 ) then ui.obj.Multiset( multiset ); end;
  
  private.Dummies_Set();
  private.Title_Set();

end;
--------------------------------------------------------------------------------
function public.Update ( quest_data )

  local id = private.Item_GetIdByName( quest_data.name );
  if  ( id ) then
  
    local multiset = {};
    
    private.Item_Update( id, quest_data, multiset );
    
    if ( #multiset > 0 ) then
    
      ui.obj.Multiset( multiset );
    
    end;
  
  end;

end;
--------------------------------------------------------------------------------
function public.Remove ( quest_name )

  local id = private.Item_GetIdByName( quest_name );
  if  ( id ) then
  
    private.Item_Delete( id );
  
  end;
  
  private.Dummies_Set();
  private.Title_Set();

end;
--------------------------------------------------------------------------------
function private.EndDragTrig ()
  
  local items_exists = #private.questlist;
  local items_needed = math.ceil( private.scc_h / private.cellsize );
  
  if ( items_exists < items_needed ) then
    ui.obj.Set( private.obj.scroller, { scrollpos = 0 } );
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.GetItemAnimName ( quest_name )

  return private.obj.cell_anim.."_"..quest_name;

end;
--------------------------------------------------------------------------------
function public.GetNameSccContainer ()
  
  return private.obj.scroller;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Item_GetIdByName ( quest_name )

  for id, quest in pairs( private.questlist ) do
  
    if ( quest.name == quest_name ) then return id; end;
  
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------
function private.Item_Create ( quest_data, multiset )

  if ( not quest_data      ) or
     ( not quest_data.name ) or
     ( ne.obj.Get( private.obj.cell_item.."_"..quest_data.name ) ) then

    return;

  end;
  
  table.insert( private.questlist, quest_data );

  ui.obj.Clone( 
    private.obj.cell_item,
    private.obj.scroller,    
    "_"..quest_data.name, 
    true
  );
  
  local cell_item = private.obj.cell_item.."_"..quest_data.name;
  local cell_anim = private.obj.cell_anim.."_"..quest_data.name;
  
  ui.obj.Set( cell_item, { pos_z = quest_data.priority }, multiset );
  ui.control.EventsSet(
    cell_anim,
    {
      func_on_click = function () private.Tap( quest_data.name ); end
    },
    multiset
  );
  
  private.Item_SetCaption(  quest_data, multiset );
  private.Item_SetTarget(   quest_data, multiset );
  private.Item_SetProgress( quest_data, multiset );
  private.Item_SetReward(   quest_data, multiset );

end;
--------------------------------------------------------------------------------
function private.Item_Update ( id, quest_data, multiset )

  if ( private.questlist[ id ] ) then
  
    private.questlist[ id ] = quest_data;
    private.Item_SetTarget(   private.questlist[ id ], multiset );
    private.Item_SetProgress( private.questlist[ id ], multiset );
  
  end;

end;
--------------------------------------------------------------------------------
function private.Item_Delete ( id )

  local quest = private.questlist[ id ];
  if  ( quest ) then
  
    ui.obj.Delete( private.obj.cell_item.."_"..quest.name );
    table.remove( private.questlist, id );
  
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Item_SetCaption ( quest_data, multiset )

  local caption    = nil;
  local stage_name = nil;
  local target     = quest_data.target;
  
  if     ( target.type == "stage" ) then stage_name = target.name;
  elseif ( target.type == "enemy" ) then stage_name = target.stage;
  end;
  
  if ( stage_name ) then
    
    local relative   = private.GetStageRelativeNumber( stage_name );
    local str_id     = "str_qst_complete_"..stage_name;
    local str_value  = StringGet( str_id, relative );
    
    if ( str_value ~= str_id ) then
      caption = str_value;
    else
      caption = StringGet( "str_qst_complete_stage", relative );
    end;
    
  else

    local pointer = quest_data.pointer;
    
    if ( pointer ) then
      local name   = ui.obj.GetNameWithoutPostfix( quest_data.name );
      local first  = private.GetStageRelativeNumber( pointer[ 1 ] );
      local second = private.GetStageRelativeNumber( pointer[ 2 ] );
      
      caption = StringGet( "str_"..name, first, second );
    else
      caption = StringGet( "str_"..quest_data.name );
    end;
  
  end;
  
  ui.obj.Set(
    private.obj.quest_anim.."_"..quest_data.name,
    { tt_description = caption },
    multiset
  );

end;
--------------------------------------------------------------------------------
function private.Item_SetTarget ( quest_data, multiset )
  
  local target_spr = private.obj.target_spr.."_"..quest_data.name;
  local target_anm = private.obj.target_anm.."_"..quest_data.name;
  local target_txt = private.obj.target_txt.."_"..quest_data.name;
  
  local anm_params, spr_params, stage_num = private.TargetParams( quest_data );
  
  if ( spr_params ) then
  
    spr_params.active  = true;
    spr_params.visible = true;
    ui.obj.Set( target_spr, spr_params, multiset );
    
    ui.obj.Disable( target_anm, true, multiset );
    ui.obj.Disable( target_txt, true, multiset );
    
  else
  
    anm_params.active  = true;
    anm_params.visible = true;
    ui.obj.Set( target_anm, anm_params, multiset );
  
    ui.obj.Disable( target_spr, true, multiset );
    
    if ( stage_num ) then
    
      ui.obj.Set(
        target_txt,
        {
          active  = true,
          visible = true,
          text    = stage_num
        },
        multiset
      );
    
    else
    
      ui.obj.Disable( target_txt, true, multiset );
    
    end;
    
  end;

end;
--------------------------------------------------------------------------------
function private.Item_SetProgress ( quest_data, multiset )

  local frame = 0;

  if ( quest_data.progress ) then
  
    local frame_max = private.frame.contract;
    if ( private.selected and private.selected == quest_data.name ) then
    
      frame_max = private.frame.expand;
    
    end;
    
    local total   = quest_data.progress.total;
    local current = quest_data.progress.current;
    frame = math.max( math.ceil( frame_max * current / total ), 0 );
  
  end;
  
  local anm_progress = private.obj.quest_anim.."_"..quest_data.name;
  ui.obj.Set( anm_progress, { frame = frame, playing = false }, multiset );

end;
--------------------------------------------------------------------------------
function private.Item_SetReward ( quest_data, multiset )

  local reward_anm = private.obj.reward_anm.."_"..quest_data.name;
  local reward_icon_anm = private.obj.reward_icon_anm.."_"..quest_data.name;
  local reward_icon_spr = private.obj.reward_icon_spr.."_"..quest_data.name;

  if ( quest_data.reward ) then
  
    ui.obj.Set( reward_anm,
      {
        active  = true,
        visible = true,
        reward  = "str_int_panel_questlist_reward"
      },
      multiset
    );
    
    if ( quest_data.reward.name == "immortality" ) then
      
      ui.obj.Disable( reward_icon_spr, true, multiset );
      ui.obj.Set( reward_icon_anm, 
      {
        animfunc = "reward",
        tt_count = ui.text.FormatTime( quest_data.reward.count ),
        rt_icon  = ui.obj.Res( quest_data.reward.type, quest_data.reward.name )
      }, multiset );
      
    else
      
      ui.obj.Disable( reward_icon_anm, true, multiset );
      ui.obj.Set( reward_icon_spr, 
      {
        res = ui.obj.Res( quest_data.reward.type, quest_data.reward.name )
      }, multiset );
      
    end;
    
  else
  
    ui.obj.Disable( reward_anm, true, multiset );
    
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Tap ( quest_name )

  local id = private.Item_GetIdByName( quest_name );
  if  ( id ) then
  
    local quest_data = private.questlist[ id ];
    local toggle     = ( #quest_data.focus.targets == 1 );
    
    if ( private.selected ) then
    
      if ( private.selected == quest_name ) then
      
        if ( toggle ) then
        
          private.Selection_Off( quest_name );
          
        else
        
          private.Selection_Set( quest_name );
        
        end;
      
      else
      
        private.Selection_Off( private.selected );
        private.Selection_On( quest_name );
      
      end;
    
    else
    
      private.Selection_On( quest_name );
    
    end;
  
  end;

end;
--------------------------------------------------------------------------------
function private.Selection_On ( quest_name )

  local id = private.Item_GetIdByName( quest_name );
  if  ( id ) then
  
    private.selected = quest_name;
    _f.Highlight_Show( private.questlist[ id ].focus );
    
    local anm_cell = private.obj.cell_anim.."_"..quest_name;
    
    ui.input.Off( anm_cell );
    ui.anim.Play( anm_cell, 
      "expand",
      function ()
        ui.input.On( anm_cell );
      end
    );
    
    local progress = private.questlist[ id ].progress;
    if  ( progress ) then
    
      local ratio = progress.current / progress.total;
      private.Animate_Progress( quest_name, ratio, true );
    
    end;
    
    local reward = private.questlist[ id ].reward;
    if  ( reward ) then
    
      private.Animate_Shift( quest_name, true );
    
    end;
  
  end;

end;
--------------------------------------------------------------------------------
function private.Selection_Set ( quest_name )

  local id = private.Item_GetIdByName( quest_name );
  if  ( id ) then
  
    _f.Highlight_Show( private.questlist[ id ].focus );
  
  end;

end;
--------------------------------------------------------------------------------
function private.Selection_Off ( quest_name )

  local id = private.Item_GetIdByName( quest_name );
  if  ( id ) then
  
    private.selected = false;
    _f.Highlight_Hide();
    
    local anm_cell = private.obj.cell_anim.."_"..quest_name;
    
    ui.input.Off( anm_cell );
    ui.anim.Play( anm_cell, 
      "contract",
      function ()
        ui.input.On( anm_cell );
      end
    );
    
    local progress = private.questlist[ id ].progress;
    if  ( progress ) then
    
      local ratio = progress.current / progress.total;
      private.Animate_Progress( quest_name, ratio, false );
    
    end;

    local reward = private.questlist[ id ].reward;
    if  ( reward ) then
    
      private.Animate_Shift( quest_name, false );
    
    end;
  
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Animate_Progress ( quest_name, ratio, expand )

  local frame =
  {
    math.ceil( ratio * private.frame.contract ),
    math.ceil( ratio * private.frame.expand   )
  };

  local anm_progress = private.obj.quest_anim.."_"..quest_name;

  ObjAnimate( anm_progress, "frame", 0, 0, "", 
    {
      0.0, 0, tern( expand )( frame[ 1 ], frame[ 2 ] ),
      0.3, 2, tern( expand )( frame[ 2 ], frame[ 1 ] )
    }
  );

end;
--------------------------------------------------------------------------------
function private.Animate_Shift ( quest_name, shift_down )

  local pos_end   = tern( shift_down )( private.cellsize, 0 );
  local cell_item = private.obj.cell_item.."_"..quest_name;
  
  local childs = ObjGetRelations( private.obj.scroller ).childs or {};
  
  local child_from;
  
  for i = 1, #childs do
  
    if ( childs[ i ] == cell_item ) then
    
      child_from = i + 1;
      break;
    
    end;
  
  end;
  
  for i = child_from, #childs do
  
    local pattern = "^"..private.obj.cell_item.."_(.*)$";
    local name = childs[ i ]:match( pattern );
    
    if ( name ) then
    
      local anm_progress = private.obj.cell_anim.."_"..name;
    
      ObjAnimate( anm_progress, "pos_y", 0, 0, "", 
        {
          0.0, 0, ui.obj.Get( anm_progress ).pos_y or 0, 
          0.3, 2, pos_end
        }
      );
    
    end;
  
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Dummies_Set ()

  local items_exists = #private.questlist;
  local items_needed = math.ceil( private.scc_h / private.cellsize );
  
  local dummies_exists = private.dummies_count;
  local dummies_needed = math.max( items_needed - items_exists, 0 );
  
  local cell_dummy = private.obj.cell_dummy.."_";
  
  if     ( dummies_exists < dummies_needed ) then
  
    for i = dummies_exists + 1, dummies_needed do
    
      local obj_name = cell_dummy..i;
    
      ui.obj.Create( obj_name, "obj" );
      ui.obj.Set( obj_name, { pos_z = 100 } );
      ui.obj.Attach( obj_name, private.obj.scroller );
    
    end;
  
  elseif ( dummies_exists > dummies_needed ) then
  
    for i = dummies_exists, dummies_needed + 1, -1 do
    
      ui.obj.Delete( cell_dummy..i );
    
    end;
  
  end;
  
  private.dummies_count = dummies_needed;

end;
--------------------------------------------------------------------------------
function private.Title_Set ()

  if ( #private.questlist < 1 ) then

    ui.obj.Set(
      "anm_int_panel_questlist",
      {
        tt_quests   = "",
        tt_noquests = private.strings.no_quests
      }
    );

  else

    ui.obj.Set(
      "anm_int_panel_questlist",
      {
        tt_quests   = private.strings.quests,
        tt_noquests = ""
      }
    );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.GetStageRelativeNumber ( stage_name )

  local stage_num = tonumber( stage_name:match( "%d+" ) or "0" );
  local relative = stage_num - ( 110 * math.floor( stage_num / 110 ) );
  relative = tern( relative == 0 )( 110, relative );

  return relative;

end;
--------------------------------------------------------------------------------
function private.TargetParams ( quest_data )

  local anm_params, spr_params, stage_num;
  
  local res_path = "assets/content/interface/"
  local res =
  {
    qst_icons = res_path.."int_panel_map_quests/anm_int_panel_quests_icons"
  };
  
  local target = quest_data.target;
  
  if ( target.type == "stage" or target.type == "stagelist" ) then
  
    anm_params = { res = res.qst_icons, animfunc = "icon_point" }; 
    stage_num  = private.GetStageRelativeNumber( target.name );
  
  elseif ( target.type == "enemy" ) then
  
    local path = "assets/content/%s/%s/icon_%s";
    spr_params =
    {
      res = path:format( target.content, target.name, target.name ),
      scale_x = 1.0,
      scale_y = 1.0
    };
  
  elseif ( target.type == "quest" or target.type == "fragment" ) then
  
    spr_params =
    {
      res     = ui.obj.Res( target.type, target.name ),
      scale_x = 0.5,
      scale_y = 0.5
    };
    
  elseif ( target.type == "resource" ) then
  
    spr_params =
    { 
      res      = ui.obj.Res( "resource", "icon_"..target.name ),
      scale_x = 1.0,
      scale_y = 1.0
    };
    
  elseif ( target.type == "build" ) then
  
    anm_params = { res = res.qst_icons, animfunc = "icon_disenchant" }; 
  
  elseif ( target.type == "mine" or target.type == "market" ) then
  
    anm_params = { res = res.qst_icons, animfunc = "icon_construct" };
    
  elseif ( target.type == "spell" ) then
  
    anm_params = { res = res.qst_icons, animfunc = "icon_spell" };
    
  else
  
    -- TODO trace
  
  end;
  
  if ( anm_params ) then
  
    anm_params.playing = true;
  
  end;
  
  return anm_params, spr_params, stage_num;

end;
--------------------------------------------------------------------------------