-- name=int_panel_worlds
--------------------------------------------------------------------------------
-- function DelegateScheme: ()
--------------------------------------------------------------------------------
local _f =
{
  ViewPort_Restore  = function ( anm_time ) end,
  ViewPort_Contract = function ( width, anm_time ) end,
  Interface_Show    = function () end,
  Interface_Hide    = function () end,
  IsContentReady    = function () end,
  GoTo              = function () end,
  OnClose           = function () end
};

local _this = {
  obj = {
    interface   = "int_panel_worlds",
    panel       = "anm_int_panel_worlds",
    body        = "anm_int_panel_worlds_body",
    scroller    = "scc_int_panel_worlds",
    transparent = "obj_int_panel_worlds_transparent",
    cell_item   = "obj_int_panel_worlds_cell",
    cell_anim   = "anm_int_panel_worlds_cell",
    cell_button = "btn_int_panel_worlds_cell_play",
    cell_dummy  = "obj_int_panel_worlds_dummy"
  },
  strings = {
    event_caption  = "str_int_panel_event_caption",
    worlds_caption = "str_int_panel_worlds_caption"
  }
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.__InitModule ()

  _this.app_o       = ui.screen.Orientation();
  _this.is_featured = false;
    
  private.scc_props = {
    full = {},
    feat = {}
  };

  private.scc_props.full.top = tern( _this.app_o == "landscape" )( 0, 80 );
  private.scc_props.feat.top = tern( _this.app_o == "landscape" )( 240, 420 );

  private.scc_props.full.h = ui.screen.Height() - private.scc_props.full.top;
  private.scc_props.feat.h = ui.screen.Height() - private.scc_props.feat.top;

  private.scc_props.full.w = 320;
  private.scc_props.feat.w = 320;

  private.scc_props.full.cellsize = 60;
  private.scc_props.feat.cellsize = 60;

  private.scc_props.full.cells_max =
    math.floor( private.scc_props.full.h / private.scc_props.full.cellsize );

  private.scc_props.feat.cells_max =
    math.floor( private.scc_props.feat.h / private.scc_props.feat.cellsize );
  
  private.panel =
  {
    width     = 320,
    height    = ui.screen.Height(),
    time_open = 0.3,
    time_hide = 0.3
  };
  
  private.section_list  = {};
  private.dummies_count = 0;
  
end;
--------------------------------------------------------------------------------
function public.__InitObjects ()
  
  local screen_w = ui.screen.Width();
  local screen_h = ui.screen.Height();
  
  ui.hub.Create( _this.obj.interface );
  
  local multiset = {};
  
  ui.obj.Set(
    _this.obj.interface,
    {
      pos_x = screen_w, 
      pos_y = 0 
    },
    multiset
  );
  
  ui.obj.Set(
    _this.obj.transparent,
    {
      pos_x          = -screen_w,
      inputrect_init = true,
      inputrect_w    = screen_w,
      inputrect_h    = screen_h,
      event_mup      = function () public.Hide(); end
    },
    multiset
  );

  ui.obj.Set(
    _this.obj.body,
    {
      tt_event       = _this.strings.event_caption,
      tt_worlds      = _this.strings.worlds_caption,
      inputrect_init = true,
      inputrect_w    = private.panel.width * 1.25,
      inputrect_h    = private.panel.height * 1.25,
    },    
    multiset
  );
  
  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Show ( is_featured )

  if ( _this.is_active ) then return; end;

  _this.is_active   = true;
  _this.is_featured = is_featured;

  local animfunc    = tern( is_featured )( "panel_event", "panel_noevent" );
  local scc_variant = tern( is_featured )( "feat", "full" );

  private.SetScroller( scc_variant );
  ui.anim.Play( _this.obj.body, animfunc );

  ui.hub.AttachContent( _this.obj.interface );
  
  ui.input.Off( _this.obj.panel );
  ui.scroller.ResetPos( _this.obj.scroller );
  --common.sound.Play( "worlds_show" );

  _f.Interface_Hide();
  _f.ViewPort_Contract( private.panel.width, private.panel.time_open );
  
  ui.anim.Play(
    _this.obj.panel,
    "appear_panel_".._this.app_o,
    function () ui.input.On( _this.obj.panel ); end,
    "static_panel_".._this.app_o
  );

end;
--------------------------------------------------------------------------------
function public.Hide ()

  if ( not _this.is_active ) then return; end;

  _this.is_active = false;

  ui.input.Off( _this.obj.panel );
  --common.sound.Play( "worlds_hide" );
  
  _f.ViewPort_Restore( private.panel.time_open );
  _f.Interface_Show();
  
  ui.anim.Play(
    _this.obj.panel,
    "disappear_panel_".._this.app_o,
    function ()
      ui.input.On( _this.obj.panel );
      ui.hub.DetachContent( _this.obj.interface );
      _f.OnClose();
    end
  );
  
  --v_log.Playhaven_LogPlacement( "worlds_panel_close" );

end;
--------------------------------------------------------------------------------
function public.IsActive ()

  return _this.is_active;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.SetList ( section_list )

  for id = #private.section_list, 1, -1 do
  
    private.Item_Delete( id );
  
  end;
  
  local multiset = {};
  
  for _, section_data in pairs( section_list ) do
  
    private.Item_Create( section_data, multiset );
  
  end;
  
  if ( #multiset > 0 ) then ui.obj.Multiset( multiset ); end;

end;
--------------------------------------------------------------------------------
function public.Add ( section_data )

  local multiset = {};

  private.Item_Create( section_data, multiset );
  
  if ( #multiset > 0 ) then ui.obj.Multiset( multiset ); end;
  
  private.SetDummies();

end;
--------------------------------------------------------------------------------
function public.Update ( section_data )

  local id = private.Item_GetIdByName( section_data.name );
  if  ( id ) then
  
    private.Item_Update( id, section_data );
  
  end;

end;
--------------------------------------------------------------------------------
function public.Remove ( section_name )

  local id = private.Item_GetIdByName( section_name );
  if  ( id ) then
  
    private.Item_Delete( id );
  
  end;
  
  private.SetDummies();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Featured_Show ()

  if ( _this.is_featured ) then return; end;

  _this.is_featured = true;

  ui.anim.Play(
    _this.obj.body,
    "panel_noevent_to_event",
    function ()
      private.SetScroller( "feat" );
    end,
    "panel_event"
  );

end;
--------------------------------------------------------------------------------
function public.Featured_Hide ()

  if ( not _this.is_featured ) then return; end;

  _this.is_featured = false;

  ui.anim.Play(
    _this.obj.body,
    "panel_event_to_noevent",
    function ()
      mediator.Notify( "Interface_Panel_Worlds_FeaturedHidden" );
    end,
    "panel_noevent"
  );

  private.SetScroller( "full" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Item_GetIdByName ( section_name )

  for id, section_data in pairs( private.section_list ) do
  
    if ( section_data.name == section_name ) then return id; end;
  
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------
function private.Item_Create ( section_data, multiset )

  if ( not section_data      ) or
     ( not section_data.name ) then

    return;

  end;

  if ( ui.obj.Exists( _this.obj.cell_item.."_"..section_data.name ) ) then

    return;

  end;
  
  table.insert( private.section_list, section_data );

  ui.obj.Clone( 
    _this.obj.cell_item,
    _this.obj.scroller,    
    "_"..section_data.name,
    1
  );

  private.Item_SetView( section_data, multiset );

end;
--------------------------------------------------------------------------------
function private.Item_Update ( id, section_data, multiset )

  if ( private.section_list[ id ] ) then
  
    private.section_list[ id ] = section_data;
    private.Item_SetView( private.section_list[ id ] );
  
  end;

end;
--------------------------------------------------------------------------------
function private.Item_Delete ( id )

  local section_data = private.section_list[ id ];
  if  ( section_data ) then
  
    ui.obj.Delete( _this.obj.cell_item.."_"..section_data.name );
    table.remove( private.section_list, id );
  
  end;

end;
--------------------------------------------------------------------------------
function private.Item_SetView ( section_data, multiset )

  local obj_name = _this.obj.cell_item.."_"..section_data.name;
  local anm_name = _this.obj.cell_anim.."_"..section_data.name;

  local world_name =
    ne.strings.Get( "str_int_dialog_worldmap_"..section_data.name );
  
  local animfunc = tern( section_data.playable )(
    "cell_world_active",
    "cell_world_inactive"
  );

  ui.obj.Set(
    obj_name, 
    {
      pos_z = math.max( 0, section_data.number )
    },
    multiset
  );
  
  ui.obj.Set(
    anm_name, 
    {
      animfunc      = animfunc,
      tt_world_name = world_name
    },
    multiset
  );

  if ( section_data.playable ) then

    ui.obj.Set(
      obj_name,
      {
        inputrect_init = true,
        inputrect_h    = private.scc_props.full.cellsize,
        inputrect_y    = private.scc_props.full.cellsize / -2,
        inputrect_w    = private.scc_props.full.w,
        inputrect_x    = private.scc_props.full.w / -2
      },
      multiset
    );

    local func_on_click = section_data.func_on_click or
    function ()
      if ( _f.IsContentReady( section_data.name ) ) then
        _f.GoTo( section_data.number );
      end;
    end;

    ui.control.EventsSet(
      obj_name,
      {
        func_on_click = func_on_click
      },
      multiset
    );

  else

    ui.obj.Set(
      obj_name,
      {
        inputrect_init = false,
        inputrect_h    = 0,
        inputrect_y    = 0,
        inputrect_w    = 0,
        inputrect_x    = 0
      },
      multiset
    );

    ui.control.EventsClear( obj_name );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.SetScroller ( variant, multiset )

  private.SetDummies();

  ui.obj.Set(
    _this.obj.scroller,
    {
      cellsize       = private.scc_props[ variant ].cellsize,
      inputrect_init = true,
      inputrect_y    = 0,
      inputrect_h    = private.scc_props[ variant ].h,
      inputrect_x    = private.scc_props[ variant ].w / -2,
      inputrect_w    = private.scc_props[ variant ].w,
      croprect_init  = true,
      croprect_y     = 0,
      croprect_h     = private.scc_props[ variant ].h,
      croprect_x     = private.scc_props[ variant ].w / -2,
      croprect_w     = private.scc_props[ variant ].w
      --enddragtrig    = private.OnDragEnd
    },
    multiset
  );  

end;
--------------------------------------------------------------------------------
function private.SetDummies ()

  local scc_variant  = tern( _this.is_featured )( "feat", "full" );
  local scc_props    = private.scc_props[ scc_variant ];
  local items_exists = #private.section_list;
  local items_needed = math.ceil( scc_props.h / scc_props.cellsize );
  
  local dummies_exists = private.dummies_count;
  local dummies_needed = math.max( items_needed - items_exists, 0 );
  
  local cell_dummy = _this.obj.cell_dummy.."_";
  
  if ( dummies_exists < dummies_needed ) then
  
    for i = dummies_exists + 1, dummies_needed do
    
      local obj_name = cell_dummy..i;
    
      ui.obj.Create( obj_name, "obj" );
      ui.obj.Set( obj_name, { pos_z = items_exists + i } );
      ui.obj.Attach( obj_name, _this.obj.scroller );
    
    end;
  
  elseif ( dummies_exists > dummies_needed ) then
  
    for i = dummies_exists, dummies_needed + 1, -1 do
    
      ui.obj.Delete( cell_dummy..i );
    
    end;
  
  end;
  
  private.dummies_count = dummies_needed;

end;
--------------------------------------------------------------------------------
function private.OnDragEnd ()
  
  local items_exists = #private.section_list;
  local scc_variant  = tern( _this.is_featured )( "feat", "full" );
  local scc_props    = private.scc_props[ scc_variant ];
  local items_needed = math.ceil( scc_props.h / scc_props.cellsize );
  
  if ( items_exists < items_needed ) then
    ui.scroller.ResetPos( _this.obj.scroller );
  end;
  
end;
--------------------------------------------------------------------------------