-- name=m_offers
--------------------------------------------------------------------------------
function public.__InitModule ()

  --
  
end;
--------------------------------------------------------------------------------
-- function ...() end;
--------------------------------------------------------------------------------
function private.GetOffers ()
  
  local offers = lib_db.Settings( "offers" );
  
  return offers;
  
end;
--------------------------------------------------------------------------------
function public.GetOffers ()
  
  local offers = private.GetOffers();
  
  return offers;
  
end;
--------------------------------------------------------------------------------
function public.Storage_GetData ()
  
  local query =
  {
    query      = "find",
    group      = "player",
    collection = "offers"
  };

  return lib_db.Execute( query ).result;
  
end;
--------------------------------------------------------------------------------
function public.Static_GetDefaultData ( type_offer )
  
  local offers = private.GetOffers().reward;
  local data_offer = {};
  
  data_offer.left_slot  = offers.left_slot[ "default_"..type_offer ];
  data_offer.right_slot = offers.right_slot[ "default_"..type_offer ];
  data_offer.center_slot = offers.center_slot[ "default_"..type_offer ];
  
  return data_offer;
  
end;
--------------------------------------------------------------------------------
function private.Storage_UpdateData ( set )
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "offers",
    set        = set
  };

  return lib_db.Execute( query ).result;
  
end;
--------------------------------------------------------------------------------
-- function ...() end;
--------------------------------------------------------------------------------
function public.CreatePlayerOffers ( boosters )
  
  local offers = private.GetOffers().reward;
  local soft_offer = private.CreateOffer( "soft", offers, boosters );
  local premium_offer = private.CreateOffer( "premium", offers, boosters );
  
  local set = {
    soft = soft_offer,
    premium = premium_offer
  };
  
  private.Storage_UpdateData( set );
  
end;
--------------------------------------------------------------------------------
function private.CreateOffer ( type_offer, offers, boosters )
  
  local data_offer = {};
  
  data_offer.left_slot  = private.GetRewardSlot( "left_slot", type_offer, offers );
  data_offer.right_slot = private.GetRewardSlot( "right_slot", type_offer, offers );
  data_offer.center_slot = private.GetRewardSlot( "center_slot", type_offer, offers, boosters );
  
  return data_offer;
  
end;
--------------------------------------------------------------------------------
function private.GetRewardSlot ( slot, type_offer, offers, boosters )
  
  local rewards = offers[ slot ][ type_offer ];
  
  if ( slot == "center_slot" ) then
    return private.GetCenterReward( rewards, boosters );
  end;
  
  local result = private.GetReward( rewards );
  local reward = {
    type = result.type,
    count = result.count,
    name = result.name
  };
  
  if ( offers[ reward.type ] and ( not reward.name ) ) then
    local item_data = private.GetReward( offers[ reward.type ] );
    reward.name = item_data.name;
  end;
  
  if ( not reward.name ) then
    reward.name = reward.type;
  end;
  
  return reward;
  
end;
--------------------------------------------------------------------------------
function private.GetCenterReward ( reward, boosters )
  
  if ( #boosters == 0 ) then
    reward.name = "hammer";
    return reward;
  end;
  
  private.CheckPurchasedLots( boosters );
  
  table.sort( boosters, function ( a, b )
    return a.period > b.period;
  end );
  
  local selected_boosters = {};
  
  table.insert( selected_boosters, boosters[ 1 ] );
  
  for i = 2, #boosters do
    if ( boosters[ i ].period < selected_boosters[ 1 ].period ) then
      break;
    else
      table.insert( selected_boosters, boosters[ i ] );
    end;
  end;
  
  table.sort( selected_boosters, function ( a, b )
    return a.priority > b.priority;
  end );
  
  reward.name = selected_boosters[ 1 ].name;
  return reward;
  
end;
--------------------------------------------------------------------------------
function private.GetReward ( rewards )
  
  local reward_random = math.random( 1, 100 );
  local reward = rewards[ 1 ];
  local chance = 0;
  
  for i = 1, #rewards do
    chance = chance + ( rewards[ i ].cnance or 0 );
    if ( reward_random <= chance ) then
      reward = rewards[ i ];
    end;
  end;
  
  return reward;
  
end;
--------------------------------------------------------------------------------
function private.CheckPurchasedLots ( boosters )
  
  local storage_data = public.Storage_GetData();
  
  if ( not storage_data.purchase_lots ) then
    return;
  end;
  
  if ( #storage_data.purchase_lots == #boosters ) then
    private.Storage_UpdateData( { purchase_lots = false } );
    return;
  end;
  
  for i = 1, #storage_data.purchase_lots do
    local purchase_booster = storage_data.purchase_lots[ i ];
    
    for j = 1, #boosters do
      if ( boosters[ j ].name == purchase_booster ) then
        table.remove( boosters, j );
        break;
      end;
    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
-- function ...() end;
--------------------------------------------------------------------------------
function public.PurchasedOffer ( offer )
  
  local booster_name = offer.center_slot.name;
  local storage_data = public.Storage_GetData();
  local time_now = TimeServerNow();
  local set = {};
  
  if ( storage_data.purchase_lots ) then
    set.purchase_lots = storage_data.purchase_lots;
    table.insert( set.purchase_lots, booster_name );
  else
    set.purchase_lots = { booster_name };
  end;
  
  set.day_start = time_now.tm_yday;
  set.premium = false;
  set.soft = false;
  
  private.Storage_UpdateData( set );
  
end;
--------------------------------------------------------------------------------
function public.IsTimeoutFinished ()
  
  local storage_data = public.Storage_GetData();
  local timeout_days = private.GetOffers().timeout_days or 30;
  local day_start = storage_data.day_start;
  
  if ( not day_start ) then
    return true;
  end;
  
  local time_now = TimeServerNow();
  local now_day = time_now.tm_yday;
  
  if ( now_day < day_start ) then
    now_day = now_day + 365;
  end;
  
  day_start = day_start + timeout_days;
  
  if ( now_day >= day_start ) then
    return true;
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------