-- name=v_globalmap
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "int_dialog_world_open",
  "int_panel_worlds"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );

  private.Init_Variables();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Init ()

  private.Init_Objects();
  private.Init_Subscribes();
  private.Worlds_SetLauncher();
  private.Worlds_SetPanel();

  public.Context_TryUnlock();

  v_interface.Scheme_AddButton( "globalmap", "right", { name = "worlds" } );
  v_interface.Scheme_AddButton( "globalmap", "top",   { name = "none" } );

end;
--------------------------------------------------------------------------------
function private.Init_Variables ()

  -- JK: ask for screen dimensions once
  private.app_h = ui.screen.Height();
  private.app_w = ui.screen.Width();
  private.app_r = ui.screen.AspectRatio();

  -- Первый запуск.
  private.first_load = true;

  private.globalmap_scale_current = 0.6;
  private.globalmap_scale_max = 1.0;
  private.globalmap_scale_min = 0.2;

  if ( private.app_h < 960 ) then

    private.globalmap_scale_current = 0.6;
    private.globalmap_scale_max = 0.75;

  end;

  private.scale_modifier = 2;
  private.offset         = 256;
  private.globalmap_w    = 2048;
  private.globalmap_h    = 1280;

  private.popup_shift = { 0, - 100 };

  private.opened = false;

  private.context_message = {
    unlocked    = false,
    triggered   = false
  };

  private.this = {
    background = "obj_globalmap_background",
    body       = "obj_globalmap_body",
    anim       = "anm_globalmap",
    marker     = "obj_globalmap_marker"
  };

  private.worldglobalmap_notification = {
    new_worlds = {}
  };

  private.world_sharing_to_facebook = "none";

end;
--------------------------------------------------------------------------------
function private.Init_Objects ()

  local screen_w = private.app_w;
  local screen_h = private.app_h;
  local scroll_h = private.scale_modifier * private.globalmap_h;

  local multiset = {};

  ui.obj.Set(
    "zmv_globalmap",
    {
      scale         = private.globalmap_scale_current,
      min_scale     = private.globalmap_scale_min,
      max_scale     = private.globalmap_scale_max,
      scrollrect_x  = 0,
      scrollrect_y  = - scroll_h,
      scrollrect_w  = private.scale_modifier * private.globalmap_w,
      scrollrect_h  = scroll_h,
      screen_w      = screen_w,
      screen_h      = screen_h,
      begindragtrig = private.DragStart,
      dragtrig      = private.DragUpdate,
      enddragtrig   = private.DragDrop
    },
    multiset
  );

  v_screen.Setup( "scr_globalmap",
  {
    func_pre_open    = private.PreOpen,
    func_open        = private.Open,
    func_close       = private.Close,
    func_after_close = private.AfterClose
  } );

  ui.obj.Set(
    private.this.background,
    {
      pos_x = 0,
      pos_y = 0
    },
    multiset
  );

  ui.obj.Set(
  private.this.anim,
    {
      pos_x =   private.offset,
      pos_y = - private.offset
    },
    multiset
  );

  ui.obj.Set(
  private.this.body,
    {
      scale_x = private.globalmap_scale_current,
      scale_y = private.globalmap_scale_current,
    },
    multiset
  );

  ui.obj.Multiset( multiset );

  ui.obj.Attach( "scr_globalmap", "ne_storage" );

end;
--------------------------------------------------------------------------------
function private.Init_Subscribes ()

  mediator.Subscribe(
    "View_Sections_EnabledList",
    private.Worlds_OnSectionsEnabled
  );

  mediator.Subscribe(
    "View_Sections_DisabledList",
    private.Worlds_OnSectionsDisabled
  );

end;
--------------------------------------------------------------------------------
function private.Worlds_OnSectionsEnabled ( enabled_list )

  if ( public.Worlds_IsPanelActive() ) then public.Worlds_HidePanel(); end;

  local multiset = {};

  for _, section_name in pairs( enabled_list ) do

    local section_number = tonumber( section_name:match( "%d+" ) );
    private.Sections_CreatePoint( { id = section_number, type = "section" }, multiset );
    private.Worlds_AddPanelItem( section_number );

  end;

  ui.obj.Multiset( multiset );

end;

--------------------------------------------------------------------------------
function private.Worlds_OnSectionsDisabled ( disabled_list )

  if ( public.Worlds_IsPanelActive() ) then public.Worlds_HidePanel(); end;

  for _, section_name in pairs( disabled_list ) do

    local section_number = tonumber( section_name:match( "%d+" ) );
    private.Sections_DestroyPoint( { id = section_number, type = "section" } );
    private.Worlds_RemovePanelItem( section_name );

  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Worlds_SetLauncher ()

  local delegates = {
    Worlds_OnTap = private.Worlds_ShowPanel
  };

  v_interface.panel.right.SetDelegates( delegates );

end;
--------------------------------------------------------------------------------
function private.Worlds_SetPanel ()

  local delegates =
  {
    ViewPort_Restore  = v_map.ViewPort_Restore,
    ViewPort_Contract = v_map.ViewPort_Contract,
    Interface_Hide    = function () v_interface.Hide(); end,
    Interface_Show    = function () v_interface.Show(); end,
    IsContentReady    = v_game.Content_IsSectionReady,

    GoTo = function ( section_number )
      int_panel_worlds.Hide();
      private.Dialog_WorldOpen_Show( { id = section_number, type = "section" } );
    end
  };

  common.delegate.Set( "int_panel_worlds", delegates );

  local section_list = v_sections.ListEnabledWithPlayableProperty();

  int_panel_worlds.SetList( section_list );

end;
--------------------------------------------------------------------------------
function private.Worlds_ShowPanel ()

  local chidren = ui.obj.Children(
    public.Worlds_GetEventPlaceholderName()
  );

  local is_featured = ( #chidren > 0 );

  if ( is_featured ) then
    mediator.Subscribe(
      "View_Minievent_FeaturedPanelEnabled",
      private.Worlds_ShowFeaturedInPanel
    );
    mediator.Subscribe(
      "View_Minievent_FeaturedPanelDisabled",
      private.Worlds_HideFeaturedInPanel
    );
    mediator.Subscribe(
      "View_Minievent_Resumed",
      private.Worlds_ShowFeaturedInPanel
    );
    mediator.Subscribe(
      "View_Minievent_Suspended",
      private.Worlds_HideFeaturedInPanel
    );
  end;

  mediator.Notify( "View_Globalmap_Worlds_PanelOpenedWithFeatured" );

  local delegates =
  {
    OnClose = function ()
      if ( is_featured ) then
        mediator.Unsubscribe(
          "View_Minievent_FeaturedPanelEnabled",
          private.Worlds_ShowFeaturedInPanel
        );
        mediator.Unsubscribe(
          "View_Minievent_FeaturedPanelDisabled",
          private.Worlds_HideFeaturedInPanel
        );
        mediator.Unsubscribe(
          "View_Minievent_Resumed",
          private.Worlds_ShowFeaturedInPanel
        );
        mediator.Unsubscribe(
          "View_Minievent_Suspended",
          private.Worlds_HideFeaturedInPanel
        );
        mediator.Notify( "View_Globalmap_Worlds_PanelClosedWithFeatured" );
      end;
    end
  };

  common.delegate.Set( "int_panel_worlds", delegates );

  int_panel_worlds.Show( is_featured );
  v_log.Playhaven_LogPlacement( "worlds_open" );

end;
--------------------------------------------------------------------------------
function private.Worlds_UpdatePanelItem ( section_name )

  local section_data = {
    number   = tonumber( section_name:match( "%d+" ) ),
    name     = section_name,
    playable = true
  };

  int_panel_worlds.Update( section_data );

end;
--------------------------------------------------------------------------------
function private.Worlds_AddPanelItem ( section_number )

  local section_data = {
    number   = section_number,
    name     = "section_"..section_number,
    playable = v_sections.Section_IsPlayable( section_number )
  };

  int_panel_worlds.Add( section_data );

end;
--------------------------------------------------------------------------------
function private.Worlds_RemovePanelItem ( section_name )

  int_panel_worlds.Remove( section_name );

end;
--------------------------------------------------------------------------------
function public.Worlds_HidePanel ( func_end )

  if ( func_end ) then
    common.delegate.Set( "int_panel_worlds", { OnClose = func_end } );
  end;

  int_panel_worlds.Hide();

end;
--------------------------------------------------------------------------------
function public.Worlds_IsPanelActive ()

  local  is_active = int_panel_worlds.IsActive();
  return is_active;

end;
--------------------------------------------------------------------------------
function public.Worlds_GetEventPlaceholderName ()

  return "obj_int_panel_worlds_event";

end;
--------------------------------------------------------------------------------
function public.Worlds_AddFeaturedPanelItem ( item_name, item_type, func_on_click )

  --JK: featured item section_number is always is -1

  local section_data = {
    number        = -1,
    name          = item_type.."_"..item_name,
    playable      = true,
    func_on_click = function ()
      int_panel_worlds.Hide();
      private.Dialog_WorldOpen_Show( { 
        id = item_name, 
        type = item_type, 
        func_on_click = func_on_click
      } );
    end
  };

  int_panel_worlds.Add( section_data );

end;
--------------------------------------------------------------------------------
function public.Worlds_RemoveFeaturedPanelItem ( item_name )

  int_panel_worlds.Remove( item_name );

end;
--------------------------------------------------------------------------------
function private.Worlds_ShowFeaturedInPanel ()

  if ( int_panel_worlds.IsActive() ) then

    int_panel_worlds.Featured_Show();

  end;

end;
--------------------------------------------------------------------------------
function private.Worlds_HideFeaturedInPanel ()

  if ( int_panel_worlds.IsActive() ) then

    int_panel_worlds.Featured_Hide();

  end;

end;
--------------------------------------------------------------------------------
function public.Worlds_AddMapIcon (
    section_id,
    section_type,
    func_on_click,
    enable_fx,
    enable_marker
  )

  local section_name = section_type.."_"..section_id;
  local section_icon = "anm_globalmap_icon_"..section_name;

  if ( not ui.obj.Exists( section_icon ) ) then
    private.Worlds_CreateMapIcon(
      section_id,
      section_type,
      func_on_click,
      enable_fx
    );
  end;

  private.Worlds_ShowMapIcon( section_id, section_type );

  if ( enable_marker ) then

    ui.anim.Play( "anm_globalmap_marker_"..section_name, "marker_active" );
  end;

end;
--------------------------------------------------------------------------------
function public.Worlds_RemoveMapIcon ( section_id, section_type )

  private.Worlds_HideMapIcon( section_id, section_type )

end;
--------------------------------------------------------------------------------
function private.Worlds_CreateMapIcon (
    section_id,
    section_type,
    func_on_click,
    enable_fx
  )

  private.Sections_CreatePoint({
    id            = section_id,
    type          = section_type,
    func_on_click = func_on_click,
    enable_fx     = enable_fx
  });

end;
--------------------------------------------------------------------------------
function private.Worlds_ShowMapIcon ( section_id, section_type )

  local section_name   = section_type.."_"..section_id;
  local section_icon   = "anm_globalmap_icon_"..section_name;
  local section_marker = "anm_globalmap_marker_"..section_name;

  ui.obj.Enable( section_icon, nil );
  ui.obj.Enable( section_marker, true );

end;
--------------------------------------------------------------------------------
function private.Worlds_HideMapIcon ( section_id, section_type )

  local section_name   = section_type.."_"..section_id;
  local section_icon   = "anm_globalmap_icon_"..section_name;
  local section_marker = "anm_globalmap_marker_"..section_name;

  ui.obj.Disable( section_icon, nil );
  ui.obj.Disable( section_marker, true );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.PreOpen ( func_after )

  local func_after_preopen = function ( is_clean_start )

    local current_section_name = v_sections.Section_NameCurrent();

    if ( is_clean_start ) then

      v_game.SilentContent_TryDownload();
      private.World_SetActiveFX( current_section_name );

      -- JK: Set player progress to current section number on Atlas
      local current_section_number = v_sections.Section_NumberCurrent();
      v_player.Progress_SetSection( current_section_number );

    end;

    public.Focus(
      "anm_globalmap_icon_"..current_section_name,
      { instant = true }
    );

    local func_minievent;

    if ( _G.v_minievent ) then

      func_minievent = v_minievent.OnMapUpdate();

    end;

    if ( func_minievent ) then

      func_minievent( func_after );

    else

      if ( func_after ) then func_after(); end;

    end;

  end;

  if ( private.first_load ) then

    private.Sections_Create( function () func_after_preopen( true ); end );
    return;

  end;

  if ( func_after_preopen ) then func_after_preopen() end;

end;
--------------------------------------------------------------------------------
function private.Open ( func_after )

  private.Music_Start();

  local is_clean_start = private.first_load;

  if ( is_clean_start ) then
    is_clean_start = v_map.IsCleanStart();
  end;
  
  local function func_update ()
    private.Update( is_clean_start );
    private.first_load = false;
    if ( func_after ) then func_after() end;
  end;
  
  if ( v_interface.window.IsAnyOnScreen() ) then
    mediator.SubscribeOnce(
      "Interface_Window_NoWindows",
      func_update
    );
  else
    func_update();
  end;

end;
--------------------------------------------------------------------------------
function private.Close ( func_after )

  private.Music_Stop();

  v_interface.Hide( { screen = "globalmap" } );

  if ( func_after ) then func_after() end;

end;
--------------------------------------------------------------------------------
function private.AfterClose ( func_after )

  if ( func_after ) then func_after() end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.SwitchToSection ( section_name )

  if ( v_sections.Section_IsEnabled( section_name )  ) and
     ( v_game.Content_IsSectionReady( section_name ) ) then

      v_sections.SetCurrentNumber( tonumber( section_name:match( "%d+" ) ) );
      v_screen.Switch( "scr_map" );

  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Sections_Create ( func_after )

  local sections_count = v_sections.GetEnabledList();

  local multiset = {};

  for _, section_name in pairs( sections_count ) do

    local section_number = tonumber( section_name:match( "%d+" ) );
    private.Sections_CreatePoint(
      { id = section_number, type = "section" },
      multiset
    );

  end;

  ui.obj.Multiset( multiset );

  if ( func_after ) then func_after(); end;

end;
--------------------------------------------------------------------------------
function private.Sections_CreatePoint ( section_data, multiset )

  local section_name   = section_data.type.."_"..section_data.id;
  local section_icon   = "anm_globalmap_icon_"..section_name;
  local section_marker = "anm_globalmap_marker_"..section_name;

  ui.obj.Clone( "anm_globalmap_icon", private.this.anim, "_"..section_name, 1 );
  ui.obj.Clone( "anm_globalmap_marker", private.this.anim, "_"..section_name, 1 );

  local res = ( section_data.type ~= "section" ) and
          "assets/content/"..section_data.type..
          "_"..section_data.id.."/globalmap/anm_icon"
          or "assets/content/globalmap/anm_icon";

  ui.obj.Set(
    section_icon,
    {
      res      = res,
      anim_tag = "at_icon"..section_data.id,
      animfunc = "icon"..section_data.id,
      playing  = true
    },
    multiset
  );

  if ( section_data.enable_fx ) then
    ui.obj.Clone(
      "anm_globalmap_icon_select_fx",
      section_icon,
      "_"..section_name
    );
    ui.obj.Set(
      "anm_globalmap_icon_select_fx_"..section_name,
      {
        res      = res,
        animfunc = "fx_select",
        playing  = true
      },
      multiset
    );
  end;

  ui.obj.Set( section_marker, {
    anim_tag = "at_marker"..section_data.id,
    playing = true
  }, multiset );

  local funcs =
  {
    func_on_click = function ()

      private.Sound_Play( "button_click" );

      ui.screen.Lock( true );

      local func_execute = function ()

        local is_enabled, is_ready, error_message;

        if ( section_data.type == "section" ) then

          is_enabled = v_sections.Section_IsEnabled( section_name );

          if ( is_enabled ) then
            is_ready = v_game.Content_IsSectionReady( section_name );
          end;

          error_message = ne.strings.Get(
            "str_int_dialog_worldmap_description_locked",
            ne.strings.Get(
              "str_int_dialog_worldmap_section_"..( section_data.id - 1 )
            )
          );

        elseif ( section_data.type == "minievent" ) then

          is_enabled    = true;
          is_ready      = true;
          error_message = "str_int_dialog_worldmap_description_unavailable";

        end;

        --DbgTrace( "is_enabled: "..tostring( is_enabled ), "is_ready: "..tostring( is_ready ) );

        if ( not is_enabled ) then
          ui.screen.Lock( false );
          ui.popup.Flat_Show(
            "str_int_dialog_worldmap_"..section_name,
            error_message,
            "obj_globalmap_icon_indicator_"..section_name,
            { 0, -80 }
          );
          return;
        end;

        if ( not is_ready ) then
          -- JK: foxy dialog is shown to download new content
          return;
        end;

        private.Dialog_WorldOpen_Show( section_data );

      end;

      public.Focus(
        "anm_globalmap_icon_"..section_name,
        {
          func_after = func_execute
        }
      );

    end
  };

  local state = "inactive";
  if ( v_sections.Section_IsEnabled( section_name ) ) then
    state = "active";
    if ( v_sections.Section_IsEnabled( v_sections.Section_NameNext( section_name ) ) ) then
      state = "complete";
    end;
  end;

  private.World_SetState( section_name, state, false );

  ui.control.EventsSet( section_icon, funcs );
  ui.control.EventsSet( section_marker, funcs );

end;
--------------------------------------------------------------------------------
function private.Sections_DestroyPoint ( section_data, multiset )

  local section_name   = section_data.type.."_"..section_data.id;
  local section_icon   = "anm_globalmap_icon_"..section_name;
  local section_marker = "anm_globalmap_marker_"..section_name;

  ui.control.EventsClear( section_icon );
  ui.control.EventsClear( section_marker );

  ui.obj.Delete( section_icon );
  ui.obj.Delete( section_marker );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Dialog_WorldOpen_Show ( section_data, func_after )

  local section_name = section_data.type.."_"..section_data.id;

  if ( section_data.type == "section" and
       not v_game.Content_IsSectionReady( section_name ) ) then
    if ( func_after ) then func_after(); end;
    return;
  end;

  local section_progress = {
    total = 0,
    done  = 0
  };
  
  if ( section_data.type == "section" ) then
    local section_stages = v_stage.Stage_GetAllBySection( section_name ) or {};
    section_progress.total = #section_stages;
    for _, stage in pairs( section_stages ) do
      local stage_data = v_stage.Stage_StorageData( stage.name ) or {};
      if ( stage_data.completed ) then
        section_progress.done = section_progress.done + 1;
      end;
    end;
  elseif ( section_data.type == "minievent" ) then
    local event_id = v_minievent.GetCurrentCustomEventId( "section" );
    section_progress.done = v_minievent.Event_GetProgress( event_id );
    section_progress.total = #v_minievent.Event_GetStages( event_id );
  end;

  local is_section_completed =
    ( section_progress.total ~=0 and
      section_progress.total == section_progress.done );

  if ( is_section_completed and section_data.type == "section" ) then
    is_section_completed =
      v_sections.Section_IsEnabled(
        v_sections.Section_NameNext( section_name )
      );
  end;

  local func_on_click = section_data.func_on_click or function ()
    public.SwitchToSection( section_name );
  end;

  local delegates = {
    Play = function ()
      local function func_OnShow ()
        int_dialog_world_open.Hide( func_on_click );
      end;
      v_interface.Loading_Show( true, func_OnShow );
    end,
    Close = function ()
      mediator.Notify( "View_Globalmap_DialogWorldOpen_Hide" );
    end,
    OnClose = function ( func_end )
      v_interface.window.Hide(
        "int_dialog_world_open",
        function ()
          if ( func_end ) then func_end(); end;
          if ( func_after ) then func_after(); end;
        end
      );
    end
  };

  v_interface.window.Show(
    "int_dialog_world_open",
    function ()
      int_dialog_world_open.Show(
        section_data.id,
        section_name,
        section_progress
      );
      mediator.Notify(
        "View_Globalmap_DialogWorldOpen_Show",
        {
          section_name         = section_name,
          is_section_completed = is_section_completed
        }
      );
    end,
    nil,
    delegates
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.ChangeList_Prepare ( change_type )

  private.changelist = private.changelist or {};

  table.insert( private.changelist, change_type );

end;
--------------------------------------------------------------------------------
function private.Update ( is_clean_start )

  --mediator.Notify( "View_Globalmap_Update_Triggered" );

  local changes = private.changelist or {};
  local update  = {}; -- callback list

  -- Notify about update started
  --local begin_msg = "View_Globalmap_Update_Started";
  --local func_notify = private.Callback_NotifyForPending( begin_msg );
  --table.insert( update, func_notify );

  -- Begin changes
  --local ch_begin_msg = "View_Globalmap_Update_ChangesBegin";
  --local func_ch_begin = private.Callback_NotifyForPending( ch_begin_msg );
  --table.insert( update, func_ch_begin );

  ui.screen.Lock( true );

  for id, change in pairs( changes ) do

    if ( change == "EnableNext" ) then
      private.Update_FocusAtCompletedWorld( update );
      private.Update_TryEnableNextWorld( update );
    end;

  end;

  --local ch_done_msg = "View_Globalmap_Update_ChangesDone";
  --local func_ch_done = private.Callback_NotifyForPending( ch_done_msg );
  --table.insert( update, func_ch_done );

  private.Update_TryShowDailyBonus( update );
  private.Update_TryShowContextMessage( update );
  private.Update_TryShowWorldOpenDialog( update, is_clean_start );

  private.Update_Finish( update );

  common.RunCallbackList( update );

  private.changelist = nil;

end;
--------------------------------------------------------------------------------
function private.Update_TryAddStep ( update, func_step, notification, params )

  if ( func_step ) then

    local func_notify =
      private.Callback_NotifyForPending( notification, params );

    -- DbgTrace( "Add to update: " .. notification, params );

    table.insert( update, func_step );
    table.insert( update, func_notify );

  end;

end;
--------------------------------------------------------------------------------
function private.Update_FocusAtCompletedWorld ( update )

  local function func_focus ( func_after )

    local section_current = v_sections.Section_NameCurrent();
    public.Focus( "anm_globalmap_icon_"..section_current,
      {
        instant    = true,
        func_after = function ()

          private.World_SetState( section_current, "complete", true, 
          function ()
            v_log.Playhaven_LogPlacement( "world_complete" );
            func_after();
          end );

        end
      }
    );

  end;

  local notification = "View_Globalmap_Update_CompletedWorldFocused";
  private.Update_TryAddStep( update, func_focus, notification );

end;
--------------------------------------------------------------------------------
function private.Update_TryEnableNextWorld ( update )

  local current_name   = v_sections.Section_NameCurrent();
  local current_number = v_sections.Section_NumberCurrent();
  local section_next   = v_sections.Section_NameNext( current_name );

  if ( v_sections.Section_IsEnabled( section_next ) ) then

    local func_enable = function ( func_after )

      v_sections.SetCurrentNumber( current_number + 1 );

      public.Focus( "anm_globalmap_icon_"..section_next,
        {
          time       = 1.0,
          func_after = function ()

            private.World_SetState( section_next, "active", true, func_after );
            private.World_SetActiveFX( section_next );
            private.Worlds_UpdatePanelItem( section_next );

          end
        }
      );

    end;

    local notification = "View_Globalmap_Update_WorldEnabled";
    private.Update_TryAddStep( update, func_enable, notification );

  end;

end;
--------------------------------------------------------------------------------
function private.Update_TryShowDailyBonus ( update )

  -- Uncomment for testing
  -- v_dailybonus.Refresh();

  local func_daily = nil;

  if ( v_dailybonus.IsReady() ) then

    func_daily = private.Dialog_DailyBonus_TryShow;

  end;

  local notification = "View_Globalmap_Update_DailyBonusDisplayed";
  private.Update_TryAddStep( update, func_daily, notification );

end;
--------------------------------------------------------------------------------
function private.Update_TryShowContextMessage ( update )

  local func_context = nil;

  if ( parts.features.IsActive( "context_message" ) ) and
     ( private.context_message.unlocked             ) and
     ( not private.context_message.triggered        ) then

    local func_show = v_game.Context_GetShowFunction( "globalmap" );

    if ( func_show ) then
      func_context = function ( func_close )
        private.Context_Show( func_show, func_close );
      end;
    end;

  end;

  local notification = "View_Globalmap_Update_ContextMessageDisplayed";
  private.Update_TryAddStep( update, func_context, notification );

  private.context_message.triggered = true;

end;
--------------------------------------------------------------------------------
function private.Update_TryShowWorldOpenDialog ( update, is_clean_start )

  local is_enabled = false;
  local context    = parts.features.GetFeatureData( "context_message" );

  for i = 1, #context do
    if ( context[ i ].name == "world_dialog" ) and
       ( context[ i ].visible                ) then
      is_enabled = true; break;
    end;
  end;

  if ( is_clean_start ) and
     ( is_enabled     ) then

    local func_ShowDialog = function ( func_after )
      local section_number = v_sections.Section_NumberCurrent();
      private.Dialog_WorldOpen_Show(
        { id = section_number, type = "section" }, func_after );
    end;

    local notification = "View_Globalmap_Update_WorldOpenDialogDisplayed";
    private.Update_TryAddStep( update, func_ShowDialog, notification );

  end;

end;
--------------------------------------------------------------------------------
function private.Update_Finish ( update )

  local function func_finish ( func_after )

    if ( not v_interface.Show( nil, func_after ) ) then
      ui.screen.Lock( false );
    end;

  end;

  local notification = "View_Globalmap_Update_Finished";
  private.Update_TryAddStep( update, func_finish, notification );

end;
--------------------------------------------------------------------------------
function private.Callback_NotifyForPending ( notification, params )

  return function ( func_after )

    mediator.Notify( notification, params );

    if ( func_after ) then

      local pending_token = public.IsPending();
      if  ( pending_token ) then

        local function func_after_pending ()

          if ( v_interface.window.IsAnyOnScreen() ) then
            v_interface.Hide();
            mediator.SubscribeOnce( "Interface_Window_NoWindows", func_after );
          else
            func_after();
          end;

        end;

        private.Pending_AddOnComplete( pending_token, func_after_pending );

      else

        func_after();

      end;

    end;

  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.World_SetState ( section_name, state, animated, func_after )

  local states_animated = {
    active = "marker_inactive_to_active",
    complete = "marker_active_to_complete"
  };
  local anm_name = "anm_globalmap_marker_"..section_name;

  if ( animated ) then

    ui.anim.Play( anm_name, states_animated[ state ], func_after, "marker_"..state );
    if ( state == "complete" ) then
      anm_name = "anm_globalmap_icon_fx_"..section_name;
      ui.anim.Play( anm_name, "fx_firework" );
    end;

    return;

  end;

  ui.anim.Play( anm_name, "marker_"..state, func_after );

end;
--------------------------------------------------------------------------------
function private.World_SetActiveFX ( section_name )

  ui.obj.Attach(
    "anm_globalmap_icon_select_fx",
    "anm_globalmap_icon_"..section_name
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Focus ( obj, params )

  -- common.DbgCallpoint( obj );

  params = params or {};

  -- NOTE: all parameters are optional:
  -- time       - focus time, default = ( distance / globalmap_w )^( 1/3 )
  -- speed      - focus speed factor, default = 1
  -- instant    - instant focus flag, default = false
  -- scale      - new scale, default = current scale
  -- func_after - after focus callback
  -- offset     - target position offset, default = { x = 0, y = 0 }

  -- Warn about breaking solid focus
  if ( IsEditor() and private.focusing_solid ) then

    local msg = "Warning! Solid focusing at '%s' breaked by '%s'";
    common.DbgCallpoint( msg:format( private.focusing_solid, obj ) );

  end;

  local func_after = params.func_after;

  private.focusing_solid = tern( func_after )( obj, false );

  local function func_focus_end ()

    if ( v_interface.IsVisible() ) then
      v_interface.Input_On();
    end

    ui.input.On( "zmv_globalmap" );

    if ( func_after ) then
      private.focusing_solid = false;
      func_after();
    end;

  end;

  -- Stop globalmap animations
  ZoomViewStopVelocity( "zmv_globalmap" );
  ObjStopAnimate( private.this.body, "pos_xy" );
  ObjStopAnimate( private.this.body, "scale_xy" );

  -- Define globalmap state changes
  local scale_now = private.Focus_ScaleCurrent();
  local scale_new = params.scale or scale_now;

  scale_new = math.min( scale_new, private.globalmap_scale_max );
  scale_new = math.max( scale_new, private.globalmap_scale_min );

  local scale_factor = scale_now / scale_new;

  local pos_now  = private.Focus_PositionCurrent();
  local offset   = params.offset;
  local pos_new  = private.Focus_PositionTarget( obj, offset, scale_new );

  local distance = private.Focus_DistanceTo( pos_now, pos_new );
  local speed    = params.speed or 1;
  local time_opt = ( distance / private.globalmap_w )^( 1/3 ) * scale_factor / speed;
  local time     = params.time or math.max( time_opt, 0.2 );

  local hold     = ( distance < 30 ) and not params.time;
  local instant  = params.instant or ( time < 1 / 25 );

  -- local tr = "\tp = %.2f,%.2f -> %.2f,%.2f; d = %.2f; "..
             -- "s = %.2f -> %.2f; t = %.2f; i = %s; h = %s";
  -- DbgTrace( tr:format(
    -- pos_now.x, pos_now.y, pos_new.x, pos_new.y, distance,
    -- scale_now, scale_new, time, instant, hold
  -- ) );

  if ( hold ) then

    pos_new = pos_now;
    instant = true;

  end;

  if ( instant ) then

    local globalmap_params =
    {
      pos_x = pos_new.x,
      pos_y = pos_new.y
    };

    if ( scale_now ~= scale_new ) then

      globalmap_params.scale_x = scale_new;
      globalmap_params.scale_y = scale_new;

    end;

    ObjSet( private.this.body, globalmap_params );
    func_focus_end();

  else

    v_interface.Input_Off();
    ui.input.Off( "zmv_globalmap" );

    if ( scale_now ~= scale_new ) then

      ObjAnimate( private.this.body, "scale_xy", 0, 0, "",
      {
        0.00, 3, scale_now, scale_now,
        time, 3, scale_new, scale_new
      } );

    end;

    ObjAnimate( private.this.body, "pos_xy", 0, 0, func_focus_end,
    {
      0.00, 3, pos_now.x, pos_now.y,
      time, 3, pos_new.x, pos_new.y
    } );

  end;

end;
--------------------------------------------------------------------------------
function private.Focus_ScaleCurrent ()

  local params = ObjGet( private.this.body ) or {};
  return params.scale_x;

end;
--------------------------------------------------------------------------------
function private.Focus_PositionCurrent ()

  local params = ObjGet( private.this.body ) or {};
  return { x = params.pos_x or 0, y = params.pos_y or 0 };

end;
--------------------------------------------------------------------------------
function private.Focus_PositionTarget ( obj, offset, scale_new )

  -- Get raw position
  local pos_raw = ui.obj.Position( obj, private.this.body );
  local pos = { x = pos_raw[ 1 ] or 0, y = pos_raw[ 2 ] or 0 };

  -- Get zoomview params
  local params = ui.obj.Get( "zmv_globalmap" );
  local scr_x, scr_y = params.screen_x, params.screen_y;
  local scr_w, scr_h = params.screen_w, params.screen_h;

  -- Shift by offset and move to screen center
  offset = offset or { x = 0, y = 0 };
  local center = { x = scr_w / 2 + scr_x, y = scr_h / 2 + scr_y };

  -- DbgTrace( ("Focus center: %s, %s"):format( center.x, center.y ) );

  pos.x = center.x + offset.x - pos.x * scale_new;
  pos.y = center.y + offset.y - pos.y * scale_new;

  -- Check bounds
  local globalmap_s        = private.scale_modifier;
  local globalmap_h, globalmap_w = private.globalmap_h, private.globalmap_w;
  -- local app_h, app_w = private.app_h, private.app_w;

  local x_min, x_max = scr_w - globalmap_s * globalmap_w * scale_new, 0;
  local y_min, y_max = scr_h,  globalmap_s * globalmap_h * scale_new;

  pos.x = math.min( pos.x, x_max );
  pos.x = math.max( pos.x, x_min );

  pos.y = math.min( pos.y, y_max );
  pos.y = math.max( pos.y, y_min );

  return pos;

end;
--------------------------------------------------------------------------------
function private.Focus_DistanceTo ( a, b )

  local dx, dy = b.x - a.x, b.y - a.y;
  return math.sqrt( dx^2 + dy^2 );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Dialog_DailyBonus_TryShow ( func_close )

  v_dailybonus.Dialog_TryShow( func_close );

end;
--------------------------------------------------------------------------------
function private.Dialog_Notifications_Show ( func_after )

  local push_reward = v_player.Playhaven_GetPushNotificationsReward();

  v_game.dialog.Notifications_Show(
    push_reward,
    function()

      return v_player.Playhaven_RegisterPushNotifications( push_reward );

    end,
    function()

      v_game.dialog.Notifications_Hide();
      if ( func_after ) then func_after() end;

    end
  );

end;
--------------------------------------------------------------------------------
function private.Share ()

  if ( v_game.IsOnline() ) then

    local icon = "icon_world_complete";
    local title = ne.strings.Get( "str_fbshare_world_complete_title" );
    local description = ne.strings.Get( "str_fbshare_world_complete_description" );

    v_game.ShareToFacebook( title, description, icon, private.ShareExecute );

  else

    v_game.SystemDialog_Show( "nointernetconnection", function () end );

  end;

end;
--------------------------------------------------------------------------------
function private.ShareExecute ()

  int_dialog_worldglobalmap.ShareComplete( private.world_sharing_to_facebook );

  private.world_sharing_to_facebook = "none";

end;
--------------------------------------------------------------------------------
function public.Worldglobalmap_SectionEnabled ()

  private.world_sharing_to_facebook = v_sections.Section_NameCurrent();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Context_TryUnlock ()

  -- Context messages conditions
  private.context_message.unlocked =
    v_stage.Stage_GetInfo( v_game.Context_GetUnlockStage() ).completed;

end;
--------------------------------------------------------------------------------
function private.Context_Show ( func_show, func_close )

  func_show( func_close );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.ViewPort_Contract ( panel_width, tme )

  ui.input.Off( "scr_globalmap" );

  ObjAnimate(
    "zmv_globalmap",
    "screen_w", 0, 0, "",
    {
      0.0, 0, ui.obj.Get( "zmv_globalmap" ).screen_w,
      tme, 0, private.app_w - panel_width
    }
  );

end;
--------------------------------------------------------------------------------
function public.ViewPort_Restore ( tme )

  ObjAnimate(
    "zmv_globalmap",
    "screen_w", 0, 0,
    function ()

      ui.input.On( "scr_globalmap" );

    end,
    {
      0.0, 0, ui.obj.Get( "zmv_globalmap" ).screen_w,
      tme, 0, private.app_w
    }
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.IsBeingDragged ()

  return private.is_dragged;

end;
--------------------------------------------------------------------------------
function private.DragStart ()

  private.is_dragged = true;

  --mediator.Notify( "View_globalmap_DragStarted" );

  ui.control.TapBreak();
  ui.popup.Cloud_Hide();
  ui.button.TryReleasePressed();

  --[[
  local tLocPosglobalmap = ObjGet( private.this.body );
  local tLocCursor = GetGameCursorPos();

  private.tPosObjglobalmap = { tLocPosglobalmap.pos_x,   tLocPosglobalmap.pos_y };
  private.tPosCursor = { tLocCursor[ 1 ], tLocCursor[ 2 ] };
  ]]

end;
--------------------------------------------------------------------------------
function private.DragUpdate ()

  --[[
  local tLocCursor = GetGameCursorPos();
  local tDelta = {
    private.tPosCursor[ 1 ] - tLocCursor[ 1 ],
    private.tPosCursor[ 2 ] - tLocCursor[ 2 ]
  };

  private.tPosCursor = { tLocCursor[ 1 ], tLocCursor[ 2 ] };

  private.tPosObjglobalmap = private.CheckPosition( { private.tPosObjglobalmap[ 1 ] - tDelta[ 1 ], private.tPosObjglobalmap[ 2 ] - tDelta[ 2 ] }, private.globalmap_scale_current );

  ObjSet( private.this.body, { pos_x = private.tPosObjglobalmap[ 1 ], pos_y = private.tPosObjglobalmap[ 2 ] } );
  ]]

  --private.GroundUpdate();

end;
--------------------------------------------------------------------------------
function private.DragDrop ()

  private.globalmap_scale_current = ObjGet( private.this.body ).scale_x;

  --mediator.Notify( "View_globalmap_DragDropped" );

  private.is_dragged = nil;

  --ObjSet( "obj_globalmap_drag", { pos_x = 0, pos_y = 0 } );

end;
--------------------------------------------------------------------------------
function private.CheckPosition ( position, scale )

  if ( position[ 1 ] > 0 ) then

    position[ 1 ] = 0;

  end;

  if ( position[ 1 ] < ( - private.scale_modifier * private.globalmap_w * scale ) + private.app_w ) then

    position[ 1 ] = ( - private.scale_modifier * private.globalmap_w * scale ) + private.app_w;

  end;

  local screen_height = private.scale_modifier * private.globalmap_h;

  if ( position[ 2 ] > screen_height ) then

    position[ 2 ] = screen_height;

  end;

  local app_height = private.app_h;

  if ( position[ 2 ] < app_height ) then

    position[ 2 ] = app_height;

  end;

  return position;

end;
--------------------------------------------------------------------------------
function private.Object_Position ( object, relative )

  relative = relative or "anm_globalmap_"..v_sections.Section_NameCurrent();

  return ui.obj.Position( object, relative );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Pending_Start ( pending_token, func_exec )

  private.Pending_Start( pending_token, func_exec );

end;
--------------------------------------------------------------------------------
function public.IsPending ()

  return private.pending;

end;
--------------------------------------------------------------------------------
function public.Pending_Complete ( pending_token )

  private.Pending_Complete( pending_token );

end;
--------------------------------------------------------------------------------
function private.Pending_Start ( pending_token, func_exec )

  -- DbgTrace( "[ VIEW ][ globalmap ] Pending started: "..pending_token );

  private.pending = pending_token;

  if ( func_exec ) then

    private.Pending_AddOnComplete( pending_token, func_exec );

  end;

end;
--------------------------------------------------------------------------------
function private.Pending_AddOnComplete ( pending_token, func_exec )

  if not ( func_exec ) then

    return;

  end;

  if not ( private.pending_oncomplete ) then

    private.pending_oncomplete = {};

  end;

  if not ( private.pending_oncomplete[ pending_token ] ) then

    private.pending_oncomplete[ pending_token ] = {};

  end;

  table.insert( private.pending_oncomplete[ pending_token ], func_exec );

end;
--------------------------------------------------------------------------------
function private.Pending_Complete ( pending_token )

  -- DbgTrace( "[ VIEW ][ globalmap ] Pending complete: "..pending_token );

  private.pending = nil;

  if ( not private.pending_oncomplete                   ) or
     ( not private.pending_oncomplete[ pending_token ]  ) or
     ( #private.pending_oncomplete[ pending_token ] < 1 ) then

     -- DbgTrace( "[ VIEW ][ globalmap ] Pending complete: nothing executed" );

    return;

  end;

  for f = 1, #private.pending_oncomplete[ pending_token ] do

    private.pending_oncomplete[ pending_token ][ f ]();

  end;

  private.pending_oncomplete[ pending_token ] = nil;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.ShowObjectPopup (
    mode,
    indicator_name,
    text,
    data,
    offset,
    functions,
    skip_focus
  )

  offset    = offset or { 0, 0 };
  functions = functions or {};

  local object_name = indicator_name:gsub( "_indicator", "" );
  local obj_pos     = private.Object_Position( object_name );

  if ( not skip_focus ) then

    local scale   = nil;
    local panel_h = 100;
    local popup_h = 240;
    local globalmap_top = -( private.globalmap_w + 2 * private.offset );

    local space = obj_pos[2] - ( globalmap_top + ( panel_h + popup_h ) / private.globalmap_scale_current );

    if ( space < popup_h + panel_h ) then

      scale = math.abs( ( popup_h / private.globalmap_scale_current ) / space );

    end;

    functions.func_scale = function( end_func )

      public.Focus(
        object_name,
        {
          scale      = scale,
          func_after = end_func,
          offset = { x = 0, y = 100 }
        }
      );

    end;

  end;

  ui.popup.Cloud_Show(
    mode,
    indicator_name,
    text,
    data,
    offset,
    1.0,
    functions
  );

end;
--------------------------------------------------------------------------------
function  public.HideObjectPopup ()

  ui.popup.Cloud_Hide();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Sound_Play ( sound_id )

  local path = "assets/content/audio/sound/aud_"..sound_id;

  common.sound.Play( path );

end;
--------------------------------------------------------------------------------
function private.Music_Start ()

  if ( not private.music ) then
    common.music.Play( "assets/content/globalmap/aud_globalmap_mus",
                    nil, nil, true );
    private.music = true;
  end;

end;
--------------------------------------------------------------------------------
function private.Music_Stop ()

  if ( private.music ) then
    common.music.Stop( "assets/content/globalmap/aud_globalmap_mus" );
    private.music = false;
  end;

end;
--------------------------------------------------------------------------------
-- function ******* DEV ******* () end
--------------------------------------------------------------------------------
function public.Cheater_Update ()

  private.Update();

end;
--------------------------------------------------------------------------------
function private.DEV_GetPointPos ()

  ObjSet( "obj_globalmap_point_DEV", { event_dragdrop = function ()

    local pos = private.Object_Position(
      "obj_globalmap_point_DEV",
      private.this.background
    );

    DbgTrace( "pos_x = "..tostring( private.scale_modifier * pos[ 1 ] ) );
    DbgTrace( "pos_y = "..tostring( private.scale_modifier * pos[ 2 ] ) );
    DbgTrace( "" );

    ObjSet( "obj_globalmap_point_DEV",
    {
      pos_x = private.scale_modifier * pos[ 1 ] + private.offset,
      pos_y = private.scale_modifier * pos[ 2 ] - private.offset
    } );

  end; } );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
--[[
  Animation format
  {
    res = "resource", -- optional
    coords = {}, -- parameters for anim_action
    name = "parent_name",
    func_preset = function () end,
    func_dummy  = function () end,
    func_action = function ( func_after ) end
  }
]]
--------------------------------------------------------------------------------
function public.Callback_Animate ( anim )

  if not ( anim ) then return false; end;

  local function func_anim ( func_after )

    if ( anim.res and anim.res ~= "" ) then

      local anm_dummy  = string.format( "anm_%s_dummy", anim.name );
      local anm_action = string.format( "anm_%s_action", anim.name );
      local anm_coords =
      {
        pos_x = anim.coords.pos_x,
        pos_y = anim.coords.pos_y,
        pos_z = 10,
        scale_x = anim.coords.scale_x or 1,
        scale_y = anim.coords.scale_y or 1,
      };

      -- Set animations
      ObjCreate( anm_dummy,  "anim" );
      ObjCreate( anm_action, "anim" );
      ObjSet(    anm_action, anm_coords );
      ObjAttach( anm_dummy,  anim.name );
      ObjAttach( anm_action, anim.name );

      ObjSet( anm_dummy,  { res = anim.res } );
      ObjSet( anm_action, { res = anim.res } );

      -- Run animations
      ui.anim.Play( anm_dummy, "dummy", function ()

        anim.func_dummy();
        ObjDelete( anm_dummy );

      end );

      local animfunc = anim.animfunc or "action";
      ui.anim.Play( anm_action, animfunc, function ()

        anim.func_action( func_after );
        ObjDelete( anm_action );

      end );

    else

      if ( anim.func_dummy ) then

        anim.func_dummy();

      end;

      anim.func_action( func_after );

    end;

  end;

  return function ( func_after )

    ui.screen.Lock( true );

    local function func_end ()
      ui.screen.Lock( false );
      if ( func_after ) then func_after(); end;
    end;

    if ( anim.func_preset ) then

      anim.func_preset();

    end;

    if ( anim.skip_focus ) then

      func_anim( func_end );

    else

      public.Focus( anim.name,
        {
          func_after = function () func_anim( func_end ) end
        }
      );

    end;

  end;

end;
--------------------------------------------------------------------------------