-- name=v_offers
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_offers",
  "int_dialog_offers"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );
  
end;
--------------------------------------------------------------------------------
function public.Init ()
  
  private.CreatePlayerOffers();
  
  v_interface.window.RegisterAsModal( "int_dialog_offers" );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.CreatePlayerOffers ()
  
  if ( not private.IsTimeoutFinished() ) then
    return;
  end;
  
  local data_boosters = v_boosters.List();
  local full_data_boosters = v_boosters.GetBoostersData( data_boosters );
  
  if ( not private.IsNeedCreatePlayerOffers( full_data_boosters ) ) then 
    return; 
  end;
  
  m_offers.CreatePlayerOffers( full_data_boosters );
  
end;
--------------------------------------------------------------------------------
function private.IsNeedCreatePlayerOffers ( data_boosters )
  
  if ( private.IsCreatePlayerOffers() ) then
    return false;
  end;
  
  local need_init = false;
  
  for i = 1, #data_boosters do
    if ( data_boosters[ i ].period > 0 ) then
      need_init = true;
      break;
    end;
  end;
  
  return need_init;
  
end;
--------------------------------------------------------------------------------
function private.IsCreatePlayerOffers ()
  
  local offers = m_offers.Storage_GetData();
  
  if ( offers.soft and offers.premium ) then
    return true;
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
function private.IsTimeoutFinished ()
  
  local is_finished = m_offers.IsTimeoutFinished();
  
  return is_finished;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.IsAvailable ()
  
  local is_finished = private.IsTimeoutFinished();
  
  return is_finished;
  
end;
--------------------------------------------------------------------------------
function public.DialogShow ( type_offer, func_OnClose )
  
  local offer = nil;
  local offers = m_offers.GetOffers();
  local price = offers.price[ type_offer ];
  
  if ( private.IsCreatePlayerOffers() ) then
    offer = m_offers.Storage_GetData()[ type_offer ];
  else
    offer = m_offers.Static_GetDefaultData( type_offer );
  end;
  
  local delegates = {
    Close = function ()
      
      private.DialogHide( func_OnClose );
      
    end,
    Buy = function ( type_offer )
      
      private.PurchaseOffer( type_offer );
      
    end,
    GetLotInfo = function ( lot_id )
      
      return v_shop.Lot_GetDataByIOSName( lot_id );
      
    end
  };
  
  v_interface.window.Show(
    "int_dialog_offers",
    function ()
      int_dialog_offers.Show( offer, type_offer, price );
    end,
    nil,
    delegates
  );
  
end;
--------------------------------------------------------------------------------
function private.DialogHide ( func_after )
  
  int_dialog_offers.Hide(
    function ()
      v_interface.window.Hide( "int_dialog_offers", func_after );
    end
  );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.PurchaseOffer ( type_offer )
  
  local offers = m_offers.GetOffers();
  
  if ( type_offer == "premium" ) then
    local lot_id = offers.price[ type_offer ].lot_id;
    v_shop.Offers_TryBuy( lot_id );
    return;
  end;
  
  if ( offers.price[ type_offer ].type == "gold" ) then
    local count = offers.price[ type_offer ].count;
    if ( v_player.Gold_TryChangeCount( -count ) ) then
      public.TryObtainOffer( type_offer );
    end;
  end;
  
end;
--------------------------------------------------------------------------------
function public.TryObtainOffer ( type_offer )
  
  local offer = nil;
  
  int_dialog_offers.ForceHide();
  
  if ( private.IsCreatePlayerOffers() ) then
    offer = m_offers.Storage_GetData()[ type_offer ];
  else
    offer = m_offers.Static_GetDefaultData( type_offer );
  end;
  
  for key, value in pairs( offer ) do
    
    value.category = "offer_"..type_offer;
    
    v_items.TryObtain( value );
    
    if ( value.type == "resource" ) then
      v_interface.Resource_Buffer_Out( value.name );
    elseif ( value.type == "gold" ) then
      v_interface.Gold_Buffer_Out();
    end;
    
  end;
  
  m_offers.PurchasedOffer( offer );
  
end;
--------------------------------------------------------------------------------